<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet
    xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:xd="http://www.oxygenxml.com/ns/doc/xsl"
    xmlns:my="file://./rules.xsl"
    exclude-result-prefixes="xs xd"
    version="2.0">
    <xsl:output method="text" />
    <xsl:strip-space elements="*" />
    <xsl:preserve-space elements="section codeph" />

<!-- Functions -->

    <xsl:function name="my:repeat" as="xs:string">
        <xsl:param name="v" as="xs:string"/>
        <xsl:param name="c"/>
        <xsl:sequence select="string-join(for $i in 1 to xs:integer($c) return $v)"/>
    </xsl:function>

    <xsl:function name="my:expand" as="xs:string">
        <xsl:param name="v" as="xs:string"/>
        <xsl:param name="c"/>
        <xsl:variable name="l" as="xs:integer" select="string-length($v)"/>
        <xsl:sequence select="concat($v,string-join(for $i in 1 to xs:integer($c) - $l return '&#x20;'))"/>
    </xsl:function>

    <xsl:function name="my:wrap" as="xs:string">
        <xsl:param name="e"/>
        <xsl:param name="w" as="xs:string"/>
        <xsl:value-of>
            <xsl:if test="$e/preceding-sibling::node() and not(matches($e/preceding-sibling::node()[1],'\s+$'))">
                <xsl:value-of select="if (matches($e,'^\s+')) then '&#x20;' else ''"/>
            </xsl:if>
            <xsl:value-of select="concat($w,normalize-space($e),$w)"/>
            <xsl:value-of select="if ($e/following-sibling::node() and matches($e,'\s+$')) then '&#x20;' else ''"/>
        </xsl:value-of>
    </xsl:function>

<!-- Generic rules -->
    
    <xsl:template match="@*|*">
        <xsl:copy>
            <xsl:apply-templates select="@*|*|text()"/>
        </xsl:copy>
    </xsl:template>
    
    <xsl:template match="text()">
        <xsl:param name="indent" select="''"/>
        <xsl:variable name="nl_indent" select="concat('&#xA;',$indent)"/>
        <xsl:variable name="prev" select="preceding-sibling::node()[1]"/>
        <xsl:variable name="tmp1" select="if (position()=1 or $prev/self::title|$prev/self::p)
                                          then replace(.,'^\s+','') else ."/>
        <xsl:variable name="tmp2" select="if (position()=last()) then replace($tmp1,'\s+$','') else $tmp1"/>
        <xsl:value-of select="replace($tmp2,'&#xA;[&#9; ]*',$nl_indent)"/>
    </xsl:template>
    
<!-- Inline markup elements -->
    
    <xsl:template match="codeph|cmdname|keyword|option">
        <xsl:param name="indent" select="''"/>
        <xsl:if test="some $attr in @* satisfies contains($attr,'  ')">
            <xsl:value-of select="concat('&#xA;',$indent)"/>
        </xsl:if>
        <xsl:value-of select="my:wrap(.,'``')"/>
    </xsl:template>
    
    <xsl:template match="b|varname|parmname">
        <xsl:value-of select="my:wrap(.,'**')"/>
    </xsl:template>
    
    <xsl:template match="i|msgph|filepath">
        <xsl:value-of select="my:wrap(.,'*')"/>
    </xsl:template>
    
    <xsl:template match="sub|sup">
        <xsl:value-of select="concat('\ :',./name(),':',my:wrap(.,'`'))"/>
    </xsl:template>

<!-- Paragraphs, lists, notes etc. -->

    <xsl:template match="p">
        <xsl:param name="indent" select="''"/>
        <xsl:if test="position()>1">
            <xsl:value-of select="concat('&#xA;',$indent)"/>
        </xsl:if>
        <xsl:apply-templates>
            <xsl:with-param name="indent" select="$indent"/>
        </xsl:apply-templates>
        <xsl:if test="last()>position()">
            <xsl:value-of select="concat('&#xA;',$indent)"/>
        </xsl:if>
    </xsl:template>
    
    <xsl:template match="note">
        <xsl:param name="indent" select="''"/>
        <xsl:if test="position()>1">
            <xsl:text>&#xA;</xsl:text>
        </xsl:if>
        <xsl:value-of select="concat($indent,'.. ',if (@type) then @type else 'note','::&#xA;&#xA;',$indent,'   ')"/> <!-- @type works for note types used in our docs -->
        <xsl:apply-templates>
            <xsl:with-param name="indent" select="concat($indent,'   ')"/>
        </xsl:apply-templates>
        <xsl:value-of select="concat('&#xA;',$indent)"/>
    </xsl:template>
        
    <xsl:template match="ul|ol">
        <xsl:param name="indent" select="''"/>
        <xsl:if test="position()>1">
            <xsl:text>&#xA;</xsl:text>
        </xsl:if>
        <xsl:for-each select="li">
            <xsl:value-of select="concat($indent,'* ')"/>
            <xsl:apply-templates>
                <xsl:with-param name="indent" select="concat($indent,'  ')"/>
            </xsl:apply-templates>
            <xsl:text>&#xA;</xsl:text>
        </xsl:for-each>
        <xsl:value-of select="$indent"/>
    </xsl:template>
    
<!-- Tables -->

    <xsl:template match="desc">
        <xsl:text>&#xA;</xsl:text>
        <xsl:apply-templates/>
        <xsl:text>&#xA;</xsl:text>
    </xsl:template>
    
    <xsl:template match="entry" mode="api">
        <xsl:apply-templates>
            <xsl:with-param name="indent" select="'  '"/>
        </xsl:apply-templates>
        <xsl:if test="position()=1">
            <xsl:text>&#xA;  \&#xA;  </xsl:text>
        </xsl:if>
    </xsl:template>

    <xsl:template match="thead" mode="api">
        <xsl:for-each select="row/entry">
            <xsl:variable name="tmp">
                <xsl:apply-templates select="node()" mode="api"/>
            </xsl:variable>
            <xsl:value-of select="normalize-space($tmp)"/>
            <xsl:if test="last()>position()">
                <xsl:text>, </xsl:text>
            </xsl:if>
        </xsl:for-each>
    </xsl:template>

    <xsl:template match="table" mode="api">
      <xsl:apply-templates select="desc"/>
        <xsl:text>&#xA;= ========================================================================================</xsl:text>
        <xsl:text>&#xA;\ </xsl:text>
      <xsl:if test="title">
        <xsl:value-of select="concat(normalize-space(title),': ')"/>
      </xsl:if>
      <xsl:apply-templates select="tgroup/thead" mode="api"/>
        <xsl:text>&#xA;==========================================================================================</xsl:text>
      <xsl:for-each select="tgroup/tbody/row">
        <xsl:text>&#xA;\ </xsl:text>
        <xsl:apply-templates select="entry" mode="api"/>
        <xsl:text>&#xA;------------------------------------------------------------------------------------------</xsl:text>
      </xsl:for-each>
        <xsl:text>&#xA;= ========================================================================================</xsl:text>
        <xsl:text>&#xA;</xsl:text>
    </xsl:template>

    <xsl:template match="table" mode="simple">
        <xsl:variable name="cols" as="xs:integer" select="xs:integer(tgroup/@cols)"/>
        <xsl:variable name="rows" as="xs:integer" select="count(tgroup//row)"/>
        <xsl:variable name="cells">
            <xsl:for-each select="tgroup//row"><row>
            <xsl:for-each select="entry[last()>position()]">
                <xsl:variable name="tmp"><xsl:apply-templates select="node()"/></xsl:variable>
                <col><xsl:value-of select="normalize-space($tmp)"/>
                    <len><xsl:value-of select="string-length(normalize-space($tmp))"/></len>
                </col>
            </xsl:for-each>
            </row></xsl:for-each>
        </xsl:variable>
        <xsl:variable name="widths" select="for $i in 1 to $cols -1 return max($cells/row/col[$i]/len), 30"/>
        <xsl:apply-templates select="desc"/>
        <xsl:if test="title">
            <xsl:text>&#xA;</xsl:text>
            <xsl:text>.. table:: </xsl:text>
            <xsl:value-of select="normalize-space(title)"/>
            <xsl:text>&#xA;&#xA;</xsl:text>
        </xsl:if>
        <xsl:variable name="indent"><xsl:if test="title"><xsl:value-of select="'   '"/></xsl:if></xsl:variable>
        <xsl:variable name="thick"><xsl:sequence select="for $i in $widths return my:repeat('=',$i)"/></xsl:variable>
        <xsl:variable name="thin"><xsl:sequence select="for $i in $widths return my:repeat('-',$i)"/></xsl:variable>
        <xsl:variable name="bigindent" select="concat($indent,my:repeat('&#x20;',$cols+sum($widths)-$widths[last()]-1))"/>

        <xsl:value-of select="concat($indent,$thick,'&#xA;')"/><!-- top border -->
        <xsl:value-of select="$indent"/>
        <xsl:for-each select="$cells/row[1]/col"><xsl:variable name="i" as="xs:integer" select="position()"/>
            <xsl:variable name="w" select="$widths[position()=$i]"/>
            <xsl:value-of select="if (text()) then my:expand(text(),$w+1) else my:expand('.',$w+1)"/>
        </xsl:for-each>
        <xsl:apply-templates select="tgroup/thead/row/entry[last()]/node()"/>
        <xsl:text>&#xA;</xsl:text>
        <xsl:value-of select="concat($indent,$thick,'&#xA;')"/><!-- header border -->

        <xsl:for-each select="tgroup/tbody/row"><xsl:variable name="r" as="xs:integer" select="position()"/>
            <xsl:value-of select="$indent"/>
            <xsl:for-each select="$cells/row[$r+1]/col"><xsl:variable name="i" as="xs:integer" select="position()"/>
                <xsl:variable name="w" select="$widths[position()=$i]"/>
                <xsl:value-of select="my:expand(text(),$w+1)"/>
            </xsl:for-each>
            <xsl:apply-templates select="entry[last()]/node()">
                <xsl:with-param name="indent" select="$bigindent"/>
            </xsl:apply-templates>
            <xsl:text>&#xA;</xsl:text>
            <xsl:if test="last()>position()">
                <xsl:value-of select="concat($indent,$thin,'&#xA;')"/><!-- bottom border -->
            </xsl:if>
        </xsl:for-each>
        <xsl:value-of select="concat($indent,$thick,'&#xA;')"/><!-- bottom border -->
    </xsl:template>

    <xsl:template match="table">
        <xsl:variable name="api-table-headers">
            <!-- If the header in the first column has one of these names, treat the table as API description -->
            <name>Pseudo-Signature</name>
            <name>Member</name>
            <name>Partitioner</name>
            <name>Buffering Policy</name>
            <name>Flag</name>
            <name>Parameter</name>
            <name>Attribute</name>
            <name>PROTECTED</name>
            <name>Entity</name>
            <name>Method</name>
        </xsl:variable>
        <xsl:variable name="hdr"><xsl:apply-templates select="tgroup/thead/row/entry[1]" mode="api"/></xsl:variable>
        <xsl:variable name="api" select="tgroup/@cols=2 and
            (some $n in $api-table-headers/name/text() satisfies contains(normalize-space($hdr),$n))"/>
        <!-- 2-column tables for API descriptions -->
        <xsl:if test="$api">
            <xsl:apply-templates select="." mode="api"/>
        </xsl:if>
        <!-- Other tables -->
        <xsl:if test="not($api)">
            <xsl:apply-templates select="." mode="simple"/>
        </xsl:if>
    </xsl:template>
    
<!-- Code, figures, images -->

    <xsl:template match="coderef">
        <xsl:text>&#xA;</xsl:text>
        <xsl:text>.. include:: </xsl:text>
        <xsl:value-of select="@href"/>
        <xsl:text>&#xA;   :code: cpp&#xA;</xsl:text>
        <xsl:text>&#xA;</xsl:text>
    </xsl:template>
    
    <xsl:template match="codeblock">
        <xsl:param name="indent" select="''"/>
        <xsl:choose>
            <xsl:when test="child::coderef">
                <xsl:apply-templates select="coderef"/>
            </xsl:when>
            <xsl:otherwise>
                <xsl:text>&#xA;</xsl:text>
                <xsl:value-of select="concat($indent,'.. code:: cpp&#xA;&#xA;',$indent,'   ')"/>
                <xsl:variable name="tmp" select="replace(.,'^&#xA;|&#xA;$','')"/>
                <xsl:value-of select="replace($tmp,'&#xA;',concat('&#xA;   ',$indent))"/>
                <xsl:value-of select="concat('&#xA;',$indent)"/>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    
    <xsl:template match="fig">
        <xsl:text>&#xA;</xsl:text>
        <xsl:text>.. figure:: </xsl:text>
        <xsl:value-of select="image/@href"/>
        <xsl:if test="image/@width">
            <xsl:value-of select="concat('&#xA;   :width: ',image/@width)"/>
        </xsl:if>
        <xsl:if test="image/@height">
            <xsl:value-of select="concat('&#xA;   :height: ',image/@height)"/>
        </xsl:if>
        <xsl:if test="image/@scale">
            <xsl:value-of select="concat('&#xA;   :align: ',image/@scale)"/>
        </xsl:if>
        <xsl:if test="image/@align">
            <xsl:value-of select="concat('&#xA;   :align: ',image/@align)"/>
        </xsl:if>
        <xsl:if test="alt">
            <xsl:value-of select="concat('&#xA;   :alt: ',alt)"/>
        </xsl:if>
        <xsl:text>&#xA;&#xA;   </xsl:text>
        <xsl:value-of select="title"/>
        <xsl:text>&#xA;</xsl:text>
    </xsl:template>

    <xsl:template match="image">
        <xsl:text>&#xA;</xsl:text>
        <xsl:text>.. image:: </xsl:text>
        <xsl:value-of select="@href"/>
        <xsl:if test="@width">
            <xsl:value-of select="concat('&#xA;   :width: ',@width)"/>
        </xsl:if>
        <xsl:if test="@height">
            <xsl:value-of select="concat('&#xA;   :height: ',@height)"/>
        </xsl:if>
        <xsl:if test="@scale">
            <xsl:value-of select="concat('&#xA;   :height: ',@scale)"/>
        </xsl:if>
        <xsl:if test="@align">
            <xsl:value-of select="concat('&#xA;   :align: ',@align)"/>
        </xsl:if>
        <xsl:text>&#xA;</xsl:text>
    </xsl:template>
    
<!-- Links -->

    <xsl:template match="related-links">
        <xsl:text>&#xA;&#xA;See also:&#xA;&#xA;</xsl:text>
        <xsl:apply-templates/>
    </xsl:template>

    <xsl:template match="link">
        <xsl:variable name="rawname"><xsl:apply-templates select="linktext/node()"/></xsl:variable>
        <xsl:variable name="linkname" select="normalize-space(replace($rawname,'`',''))"/>
        <xsl:text>* :doc:`</xsl:text>
        <xsl:value-of select="$linkname"/>
        <xsl:text> &lt;</xsl:text>
        <xsl:value-of select="replace(@href,'\.xml.*$','')"/>
        <xsl:text>&gt;`&#xA;</xsl:text>
    </xsl:template>

    <xsl:template match="xref">
        <xsl:variable name="rawname"><xsl:apply-templates select="node()"/></xsl:variable>
        <xsl:variable name="linkname" select="normalize-space(replace($rawname,'`',''))"/>
        <xsl:choose>
            <xsl:when test="@scope='local'">
                <xsl:text>:doc:`</xsl:text>
                <xsl:value-of select="$linkname"/>
                <xsl:text> &lt;</xsl:text>
                <xsl:value-of select="replace(@href,'\.xml.*$','')"/>
                <xsl:text>&gt;`</xsl:text>
            </xsl:when>
            <xsl:otherwise>
                <xsl:text>`</xsl:text>
                <xsl:value-of select="$linkname"/>
                <xsl:text> &lt;</xsl:text>
                <xsl:value-of select="@href"/>
                <xsl:text>&gt;`_</xsl:text>
            </xsl:otherwise>
        </xsl:choose>    
    </xsl:template>
    
<!-- Document structure elements -->

    <xsl:template match="metadata">
    </xsl:template>

    <xsl:template match="title">
        <xsl:variable name="title_text" select="normalize-space(.)"/>
        <xsl:variable name="title_len" select="string-length($title_text)"/>
        <xsl:if test="not(following-sibling::body|following-sibling::refbody)">
            <!-- the main title is managed by higher-level stylesheets -->
            <xsl:value-of select="concat('&#xA;&#xA;',$title_text,'&#xA;',my:repeat('-',$title_len))"/>
            <xsl:text>&#xA;</xsl:text>
        </xsl:if>
    </xsl:template>
    
<!--
<xsl:template match="refbody">
<xsl:variable name="test">Put here an XML sample to check</xsl:variable>
<xsl:variable name="out"><xsl:apply-templates select="$test"/></xsl:variable>
<xsl:message>test: !<xsl:value-of select="$out"/>!</xsl:message>
</xsl:template>
-->

</xsl:stylesheet>
